package mrl.common;

import javolution.util.FastList;
import javolution.util.FastSet;
import math.geom2d.Point2D;
import math.geom2d.conic.Circle2D;
import mrl.world.MrlWorld;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

/**
 * @author vahid hooshangi
 */


/**
 * in this class check the agent is stuck or not, if agent try move to 50% of targets and failed them, agent is in stuck
 */
public class AmIStuckChecker {
    private MrlWorld world;
    private List<EntityID> targets;
    private int stuckThreshold;
    private int randomCount=10;
    Random random;


    public AmIStuckChecker(MrlWorld world, int stuckThreshold, int randomCount) {
        targets = new FastList<EntityID>();
        this.stuckThreshold = stuckThreshold;
        this.world = world;
        this.random = new Random(System.currentTimeMillis());

    }

    /**
     * this function check agent is Stuck or not, and if agent is on blockage return tru
     *
     * @param x integer representing the x position of the agent in the map
     * @param y integer representing the y position of the agent in the map
     * @return true if the agent is stuck false otherwise
     */
    public boolean amIStuck(int x, int y) {
        if (!Util.isOnBlockade(world)) {
            buildCircles(x, y);
            if (targets.size() > 0) {

                return checkStatus(choiceRandomTargets());
            }
        }

        return true;
    }

    private Set<EntityID> choiceRandomTargets() {
        Set<EntityID> randomTargets = new FastSet<EntityID>();
        int rand;
        int count = 0;
        while (count < randomCount) {
            rand = random.nextInt(targets.size());
            if (randomTargets.contains(targets.get(rand))) {
                continue;
            }
            randomTargets.add(targets.get(rand));
            count++;
        }

        return randomTargets;
    }



    /**
     * this function create two circles, first is inner circle that centre is human(x,y) and radius viewDistance
     * second is outer circle that centre is human(x,y) and radius is viewDistance * 1.5
     *
     * @param x
     * @param y
     */
    private void buildCircles(int x, int y) {
        Circle2D innerCircle = new Circle2D(new Point2D(x, y), world.getPlatoonAgent().viewDistance);
        Circle2D outerCircle = new Circle2D(new Point2D(x, y), world.getPlatoonAgent().viewDistance * 1.5);

        findPoints(innerCircle, outerCircle);
    }

    /**
     * findPoints functions find all of entities that in area between innerCircle and outerCircle
     *
     * @param innerCircle the Circle with agent(x, y) and viewDistance radius
     * @param outerCircle the Circle with agent(x, y) and 1.5 * viewDistance radius
     */
    private void findPoints(Circle2D innerCircle, Circle2D outerCircle) {

        for (StandardEntity se : world.getBuildings()) {
            if (se instanceof Building
                    && contains(outerCircle, ((Building) se).getX(), ((Building) se).getY())
                    && !contains(innerCircle, ((Building) se).getX(), ((Building) se).getY())) {
                targets.add(se.getID());
            }
        }

        for (StandardEntity se : world.getRoads()) {
            if (se instanceof Road
                    && contains(outerCircle, ((Road) se).getX(), ((Road) se).getY())
                    && !contains(innerCircle, ((Road) se).getX(), ((Road) se).getY())) {
                targets.add(se.getID());
            }
        }
    }


    /**
     * In this function we send move act to all of @targets and
     * if more than stuckThreshold move act is failed, this human is in stuck else agent is free
     *
     * @param entityIDs , roads and buildings entityIds that choice random from targets
     * @return boolean true if the agent is stuck
     */
    private boolean checkStatus(Set<EntityID> entityIDs) {

        if(entityIDs == null || entityIDs.isEmpty()){
            return false;
        }

        int unSuccessfulMove = 0;

        StandardEntity se;
        Area source = (Area) world.getSelfPosition();
        for (EntityID id : entityIDs) {
            se = world.getEntity(id);
            if (se instanceof Area) {
                if (world.getPlatoonAgent().getPathPlanner().planMove(source, (Area) se, 0, false).isEmpty()) {
                    unSuccessfulMove++;
                }
            }
        }


        if (((unSuccessfulMove / entityIDs.size()) * 100) > stuckThreshold) {
            return true;
        }
        return false;
    }

    private boolean contains(Circle2D circle2D, double x, double y) {
        double dx = circle2D.getCenter().getX() - x;
        double dy = circle2D.getCenter().getY() - y;

        return ((dx * dx) + (dy * dy)) < circle2D.getRadius() * circle2D.getRadius();
    }

}
