package mrl.partitioning;

import javolution.util.FastMap;
import mrl.ambulance.marketLearnerStrategy.AmbulanceUtilities;
import mrl.assignment.GreedyAssignment;
import mrl.assignment.IAssignment;
import mrl.common.MRLConstants;
import mrl.common.TimestampThreadLogger;
import mrl.partitioning.costMatrixMaker.CostMatrixMaker;
import mrl.partitioning.segmentation.SegmentType;
import mrl.viewer.layers.MrlClusterLayer;
import mrl.viewer.layers.MrlPartitionsLayer;
import mrl.viewer.layers.MrlRendezvousLayer;
import mrl.viewer.layers.MrlSubClusterLayer;
import mrl.world.MrlWorld;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

/**
 * Author: Pooya Deldar Gohardani
 * Date: 2/12/12
 * Time: 5:27 PM
 */
public class DefaultAmbulancePartitionManager implements IManager {
    private static final Log logger = LogFactory.getLog(DefaultAmbulancePartitionManager.class);

    private MrlWorld world;
    private AmbulanceUtilities utilities;
    private PartitionHelper partitionHelper;
    private SegmentType segmentType;

    private List<Partition> partitions;
    private PartitionUtilities partitionUtilities;
    private CostMatrixMaker costMatrixMaker;
    private IAssignment assignmentMethod;
    private Map<Human, Partition> humanPartitionMap;
    private Map<Human, Partition> previousHumanPartitionMap;
    private final int NUMBER_OF_AGENTS_IN_PARTITION = 1;

    //Map of PartitionID to the list of nearest refuge path, which contains in that partition
    private Map<EntityID, Set<EntityID>> partitionsRefugePathMap;

    //TODO @BrainX Make this singleton, with Spring maybe?

    /**
     * Creates an instance of DefaultManager
     *
     * @param world     The world this instance is supposed to perform partitioning operations.
     * @param utilities reference to a {@link mrl.police.moa.PoliceForceUtilities}
     */
    public DefaultAmbulancePartitionManager(MrlWorld world, AmbulanceUtilities utilities) {
        this.world = world;
        this.utilities = utilities;
    }

    private void initialVariables(MrlWorld world) {

        segmentType = SegmentType.ENTITY_CLUSTER;
        this.partitionHelper = new PartitionHelper(world);

        partitionUtilities = new PartitionUtilities(world);
        costMatrixMaker = new CostMatrixMaker(world);

        if (humanPartitionMap == null) {
            previousHumanPartitionMap = new FastMap<Human, Partition>();
        } else {
            previousHumanPartitionMap.putAll(humanPartitionMap);
        }

        partitions = new ArrayList<Partition>();
        humanPartitionMap = new FastMap<Human, Partition>();
        partitionsRefugePathMap = new FastMap<EntityID, Set<EntityID>>();

        // TODO @BrainX Hungarian implementation assigns a single agent to multiple partitions
        assignmentMethod = new GreedyAssignment();
//        assignmentMethod = new HungarianAssignment();
    }


    /**
     * {@inheritDoc}
     */
    @Override
    public void initialise() {

        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Entering initialize...");

        initialVariables(world);


        Partition worldPartition = partitionHelper.makeWorldPartition();

        List<StandardEntity> ambulanceTeams = new ArrayList<StandardEntity>(world.getAmbulanceTeamList());
        int numberOfAmbulances = ambulanceTeams.size();
        //finding number of needed Segments
        int numberOfSegments = numberOfAmbulances / NUMBER_OF_AGENTS_IN_PARTITION;
        if (numberOfSegments == 0) {
            numberOfSegments = 1;
        }

        //segment world into specified number of partitions
        partitions = split(worldPartition, numberOfSegments);

        //updating values of each constructed partitions
        updatePartitions(ambulanceTeams.size());

        //handles assignment of each agent to a proper partition
        updateAssignment(ambulanceTeams);


        updateViewer();


        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Returning from initialize.");
        TimestampThreadLogger.getCurrentThreadLogger().flush();
    }

    private void updateViewer() {
        if (MRLConstants.LAUNCH_VIEWER) {
            for (Partition partition : getPartitions()) {
                if (partition.getPolygon() == null) {
                    partition.computeConvexHull();
                }
            }

            MrlClusterLayer.PARTITIONS_MAP.put(world.getSelf().getID(), getPartitions());
            MrlClusterLayer.HUMAN_PARTITION_MAP.put(world.getSelf().getID(), findHumanPartition(world.getSelfHuman()));
            if (findHumanPartition(world.getSelfHuman()) == null) {
                if (findHumanPartition(world.getSelfHuman()) == null) {
                    System.out.println(world.getSelf().getID() + " I have no partition in time :" + 0);
//        }
                } else {

                    System.out.println(world.getSelf().getID() + " my Partition is:" + findHumanPartition(world.getSelfHuman()).getId() + " NumberOfNeeded:" + findHumanPartition(world.getSelfHuman()).getNumberOfNeededPFs());
                }
            }
            MrlSubClusterLayer.Partitions = getPartitions();
            MrlPartitionsLayer.PARTITIONS = getPartitions();
            MrlRendezvousLayer.partitions = getPartitions();
        }

    }


    /**
     * {@inheritDoc}
     */
    @Override
    public Partition findPartitionAtArea(EntityID areaId) {
        Partition foundPartition = null;
        for (Partition partition : partitions) {
            if (partition.containsInEntities(areaId)) {
                foundPartition = partition;
                break;
            }
        }
        return foundPartition;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void forceAssignAgent(Human platoonAgent, Partition partition) {
        humanPartitionMap.put(platoonAgent, partition);
    }

    @Override
    public ChangeSet preprocessChanges(ChangeSet changeSet) {
        return null;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Partition> split(Partition partition, int n) {
        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Entering split...");
        List<Partition> splitPartitions = partitionHelper.split(partition, n, segmentType);
        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Returning from split.");
        return splitPartitions;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void updatePartitions(int agents) {
        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Entering updatePartitions...");

        //update independent features
        for (Partition partition : partitions) {
            updatePartition(partition);
        }

        // compute Number Of Needed AmbulanceTeams for each partition. It is Considered to allocate the same number of
        // agents to each partition because the size of partitions are close to each other
        if (agents > 0) {
            //update dependent features
            int remainedAgents = agents;
            int i = 0;

            int assigned = 0;
            while (remainedAgents > 0) {
                assigned = partitions.get(i).getNumberOfNeededATs();
                partitions.get(i).setNumberOfNeededATs(assigned + 1);
                remainedAgents--;
                i++;
                if (i == partitions.size()) {
                    i = 0;
                }
            }
        }

        if (partitionsRefugePathMap == null || partitionsRefugePathMap.isEmpty()) {
            partitionsRefugePathMap = new FastMap<EntityID, Set<EntityID>>();
            partitionsRefugePathMap.putAll(partitionUtilities.findPartitionsRefugePaths(partitions));
            for (Partition partition : partitions) {
                partition.setRefugePathsToClearInPartition(partitionUtilities.findRefugePathsToClearInPartition(partition, partitionsRefugePathMap));
            }
        }


        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Returning from updatePartitions.");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void updateAssignment(List<StandardEntity> agents) {
        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Entering updateAssignment...");

        Collections.sort(partitions, Partition.PARTITION_NEEDED_AT_COMPARATOR);

        List<Partition> deflatedPartitions = arrangePartitions(partitions);

        double[][] costMatrix;
        if (previousHumanPartitionMap.isEmpty()) {
            costMatrix = costMatrixMaker.makingCostMatrix(deflatedPartitions, agents, this.humanPartitionMap, false);
        } else {
            this.humanPartitionMap.clear();
            this.humanPartitionMap.putAll(previousHumanPartitionMap);
            costMatrix = costMatrixMaker.makingCostMatrix(deflatedPartitions, agents, this.humanPartitionMap, true);

        }
        int[][] assignment = assignmentMethod.computeAssignments(costMatrix);

        for (int i = 0; i < assignment.length; i++) {
            this.humanPartitionMap.put((Human) agents.get(assignment[i][0]), deflatedPartitions.get(i));
//            logger.debug(world.getTime() + " self: " + world.getSelf().getID() + " agentID: " + agents.get(assignment[i][0]).getID() + " partition: " + deflatedPartitions.get(i).getId());
//            System.out.println((world.getTime() + " self: " + world.getSelf().getID() + " agentID: " + agents.get(assignment[i][0]).getID() + " partition: " + deflatedPartitions.get(i).getId()));
        }
        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Returning from updateAssignment.");
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Partition findHumanPartition(Human human) {
        return humanPartitionMap.get(human);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<Partition> getPartitions() {
        return partitions;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Partition findPartition(EntityID partitionID) {
        Partition foundPartition = null;
        for (Partition partition : partitions) {
            if (partition.getId().equals(partitionID)) {
                foundPartition = partition;
                break;
            }
        }
        return foundPartition;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Set<Map.Entry<Human, Partition>> getHumanPartitionEntrySet() {
        return humanPartitionMap == null ? null : humanPartitionMap.entrySet();
    }

    /**
     * Updates partition properties
     *
     * @param partition partition to update
     */
    private void updatePartition(Partition partition) {
        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Entering updatePartition...");
        if (!partition.isDead()) {
            partition.setBurningBuildings(partitionUtilities.getBurningBuildings(partition));
//            partition.setUnVisitedBuilding(partitionUtilities.getUnVisitedBuildings(partition));
//            partition.setBlockedAgents(partitionUtilities.getBlockedAgents(partition));
//            partition.setBuriedAgents(partitionUtilities.getBuriedAgents(partition));
            partition.setDead(partitionUtilities.isPartitionDead(partition));
        }
        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Returning from updatePartition.");
    }


    /**
     * Creates a flat representation of the specified {@code partitions} and their one level deep children.
     * <b>Note:</b> This method does not support multilevel partitions. Only one level of children is supported.
     *
     * @param partitions partitions to deflate
     * @return A list of parameters without children
     */
    private List<Partition> arrangePartitions(List<Partition> partitions) {
        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Entering arrangePartitions...");

        List<Partition> allPartitions = new ArrayList<Partition>();


        for (Partition partition : partitions) {
            for (int i = 0; i < partition.getNumberOfNeededATs(); i++) {
                allPartitions.add(partition);
            }
        }
        TimestampThreadLogger.getCurrentThreadLogger().log(world.getSelf().getID() + " Returning from arrangePartitions.");
        return allPartitions;
    }

}
