package mrl.partitioning.costMatrixMaker;

import javolution.util.FastMap;
import mrl.common.Util;
import mrl.partitioning.Partition;
import mrl.world.MrlWorld;
import mrl.world.routing.pathPlanner.IPathPlanner;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;

import java.util.List;
import java.util.Map;
import java.util.Set;

import static java.lang.Integer.MAX_VALUE;

/**
 * Author: Pooya Deldar Gohardani
 * Date: 5/9/12
 * Time: 7:32 PM
 */
public class CostMatrixMaker {


    private MrlWorld world;

    public CostMatrixMaker(MrlWorld world) {
        this.world = world;
    }

    /**
     * This method makes cost matrix from each agent distances to center of each partition(first for previously assigned ones)
     *
     * @param deflatedPartitions
     * @param agentList          list of agents to compute their costs
     * @param humanPartitionMap  map of assigned humans to partitions  @return a two dimension matrix of costs
     * @return
     */
    public double[][] makingCostMatrix(List<Partition> deflatedPartitions, List<StandardEntity> agentList, Map<Human, Partition> humanPartitionMap, boolean isNewAssignment) {
        int n = agentList.size();
        double[][] costMatrix = new double[n][n];
        Partition partition;
        int previouslyAssigned = 0;
        Map<Integer, Integer> assignedPartitions = new FastMap<Integer, Integer>();

        if (!isNewAssignment) {

            for (int i = 0; i < n; i++) {
                partition = humanPartitionMap.get((Human) agentList.get(i));
                if (partition != null) {
                    for (int j = 0; j < n; j++) {
                        previouslyAssigned = 0;
                        // if agent i had any assigned partition
                        // cost for agent i to reach current partition is zero and to others is Max value to keep each assigned agent
                        // in its current partition
                        if (deflatedPartitions.get(j).equals(partition)) {
                            costMatrix[i][j] = 0;
                            if (assignedPartitions.containsKey(j)) {
                                previouslyAssigned = assignedPartitions.get(j);
                            }
                            previouslyAssigned += 1;
                            assignedPartitions.put(j, previouslyAssigned);
                        } else {
                            costMatrix[i][j] = MAX_VALUE;
                        }
                    }
                }
            }


            for (int i = 0; i < n; i++) {
                partition = humanPartitionMap.get((Human) agentList.get(i));
                for (int j = 0; j < n; j++) {
                    // if agent i had no assigned partition
                    if (partition == null) {
//                    if (!assignedPartitions.contains(j)) {
                        if (assignedPartitions.containsKey(j) && assignedPartitions.get(j) < deflatedPartitions.get(j).findNumberOfNeededAgents(agentList.get(i))) {
                            costMatrix[i][j] = computeDistance(agentList.get(i), deflatedPartitions.get(j));
                        } else if (!assignedPartitions.containsKey(j) && deflatedPartitions.get(j).findNumberOfNeededAgents(agentList.get(i)) > 0) {
                            costMatrix[i][j] = computeDistance(agentList.get(i), deflatedPartitions.get(j));
                        } else {
                            costMatrix[i][j] = MAX_VALUE;
                        }
                    }
                }
            }

        } else {
            for (int i = 0; i < n; i++) {
                partition = humanPartitionMap.get(agentList.get(i));
                for (int j = 0; j < n; j++) {
                    costMatrix[i][j] = Util.distance(partition.getCenter(), deflatedPartitions.get(j).getCenter());
                }
            }

        }

        return costMatrix;
    }


    public double[][] makingCostMatrix(List<Partition> deflatedPartitions, List<StandardEntity> agentList, Set<Map.Entry<Human, Partition>> humanPartitionMap) {
        int n = agentList.size();
        double[][] costMatrix = new double[n][n];
        Partition partition;
        int previouslyAssigned = 0;
        Map<Integer, Integer> assignedPartitions = new FastMap<Integer, Integer>();


        for (int i = 0; i < n; i++) {
            partition = findPartition(humanPartitionMap, agentList.get(i));
            for (int j = 0; j < n; j++) {
                costMatrix[i][j] = Util.distance(partition.getCenter(), deflatedPartitions.get(j).getCenter());
            }
        }

        return costMatrix;
    }

    private Partition findPartition(Set<Map.Entry<Human, Partition>> humanPartitionMap, StandardEntity agentEntity) {
        Partition foundPartition = null;
        for (Map.Entry<Human, Partition> humanPartitionEntry : humanPartitionMap) {
            if (humanPartitionEntry.getKey().getID().equals(agentEntity.getID())) {
                foundPartition = humanPartitionEntry.getValue();
                break;
            }
        }
        return foundPartition;
    }


    public int[][] makingCostMatrix_BasedOnID(List<StandardEntity> agentList) {

        int n = agentList.size();
        int[][] costMatrix = new int[n][n];
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                // cost for agent i to reach current partition is zero and to others is Max value to keep each assigned agent
                // in its current partition
                costMatrix[i][j] = agentList.get(i).getID().getValue();
            }
        }


        return costMatrix;

    }


    /**
     * Computes simple euclidean distance between a policeForce and center of a partition
     *
     * @param agent     agent to compute its distance
     * @param partition a partition of the map
     * @return an integer which shows the distance of two entities
     */
    private int computeDistance(StandardEntity agent, Partition partition) {
        IPathPlanner pathPlanner = world.getPlatoonAgent().getPathPlanner();
        Human human = (Human) agent;
        pathPlanner.planMove((Area) world.getEntity(human.getPosition()), (Area) partition.getCenterEntity(), 0, true);
        return pathPlanner.getPathCost();
//        return Util.distance(agent.getLocation(world), partition.getCenter());
    }


}
