package mrl.police;

import mrl.common.CommandException;
import mrl.common.MRLConstants;
import mrl.common.TimeOutException;
import mrl.common.Util;
import mrl.common.comparator.ConstantComparators;
import mrl.communication.Packet;
import mrl.helper.RoadHelper;
import mrl.partitioning.DefaultManager;
import mrl.partitioning.IManager;
import mrl.partitioning.Partition;
import mrl.platoon.MrlPlatoonAgent;
import mrl.platoon.State;
import mrl.platoon.genericsearch.*;
import mrl.platoon.search.SearchHelper;
import mrl.police.moa.IAuction;
import mrl.police.moa.PoliceForceUtilities;
import mrl.police.moa.Target;
import mrl.police.strategies.PolicePrioritizationStrategy;
import mrl.police.strategies.PoliceStrategy;
import mrl.task.PoliceActionStyle;
import mrl.task.Task;
import mrl.viewer.layers.MrlClusterLayer;
import mrl.viewer.layers.MrlPartitionsLayer;
import mrl.viewer.layers.MrlRendezvousLayer;
import mrl.viewer.layers.MrlSubClusterLayer;
import mrl.world.object.Entrance;
import mrl.world.routing.path.Path;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import rescuecore2.messages.Command;
import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;

import java.util.*;


/**
 * MRL police force agent.
 * <p/>
 * Since 1.1: Added multiple Operation Mode Support
 *
 * @author Legacy
 * @author BrainX
 * @version 1.1
 */
public class MrlPoliceForce extends MrlPlatoonAgent<PoliceForce> {
    private static final Log logger = LogFactory.getLog(MrlPoliceForce.class);
    private static final int SECOND_PARTITIONING_TIME = 9;
    // TODO @BrainX Clean-up

    /**
     * Operation Mode of any {@link MrlPoliceForce}.
     *
     * @since 1.1
     */
    private static enum OperationMode {
        /**
         * In some cases, Police Forces perform partitioning for those of them on the road. Later, if anyone else
         * recovers or gets out on the road, they will be assigned to "Bigger" partitions (Those who have more buildings).
         */
        SINGULAR_PARTITIONING,
        /**
         * In some cases, Police Forces perform partitioning for those of them on the road. Later, after a few cycles (10)
         * they will perform partitioning again based on all those on the road or were inside building and are now on the road.
         * <br/><b>Note: </b>After first step, this Operation Mode is similar the SINGULAR_PARTITIONING and developers are
         * encouraged to reuse the code.
         */
        DUAL_PARTITIONING,
    }


    protected MrlPoliceForceWorld world;
    protected ClearHereHelper clearHereHelper;
    protected PoliceMessageHelper policeMessageHelper;
    protected Path targetPath;
    //    protected PoliceForceBid currentTask;
    private Road targetRoad = null;
    private Collection<EntityID> unexploredBuildings;


    private IAuction auction;
    private PoliceForceUtilities utilities;
    private Task myTask;
    private PoliceConditionChecker pcc;
    //    List<IBehavior> myBehaviors;
//    private AbstractBehaviorHelper policeBehaviorsHelper;
    private boolean inPartition = true;
    private PoliceStrategyTypes strategyTypes;
    private PoliceStrategy strategy;
    private Human selfHuman;
    private List<EntityID> pathToGo;
    private ClearHelper clearHelper;
    private IManager partitionManager;
    private SearchHelper searchHelper;

    /**
     * Operation Mode of this {@link MrlPoliceForce}
     *
     * @see OperationMode
     * @since 1.1
     */
    private OperationMode operationMode;

    @Override
    public String toString() {
        return "MRL police force ID: " + this.getID().getValue();
    }

    @Override
    protected void postConnect() {
        super.postConnect();
        this.world = (MrlPoliceForceWorld) super.world;
        model.indexClass(StandardEntityURN.ROAD);
        clearHereHelper = new ClearHereHelper(world);//old clear helper version
        clearHelper = new ClearHelper(world);
        policeMessageHelper = new PoliceMessageHelper(world, this, messageManager);
        Collections.sort(world.getPoliceForceList(), ConstantComparators.ID_COMPARATOR);

//        unexploredBuildings = new HashSet<EntityID>(getMyPartition().getBuildingsID());

        utilities = new PoliceForceUtilities(world, config);
//        auction = new Auction(world,this,policeMessageHelper);
//        auction.setTargetBidsMap(world.getTargetBidsMap());
//        auction.setAvailableAgents(world.getPoliceForceList().size());
        pcc = new PoliceConditionChecker(world, utilities);
//        behaviorHelper = new PoliceBehaviorsHelper(world);
//        myBehaviors = behaviorHelper.getBehaviors();
//        policeBehaviorsHelper = new PoliceBehaviorsHelper(world, this, config, policeMessageHelper);
        selfHuman = world.getSelfHuman();


        possibleBuildingSearchDecisionMaker = new PossibleBuildingSearchDecisionMaker(world);
        possibleBuildingSearchDecisionMaker.setSearchInPartition(true);
        heatTracerDecisionMaker = new HeatTracerDecisionMaker(world);
        heatTracerDecisionMaker.setSearchInPartition(true);
        stupidSearchDecisionMaker = new StupidSearchDecisionMaker(world);
        stupidSearchDecisionMaker.setSearchInPartition(true);

        possibleBuildingSearchManager = new PossibleBuildingSearchManager(world, this, possibleBuildingSearchDecisionMaker, defaultSearchStrategy);
        heatTracerSearchManager = new HeatTracerSearchManager(world, this, heatTracerDecisionMaker, defaultSearchStrategy);
        stupidSearchManager = new StupidSearchManager(world, this, stupidSearchDecisionMaker, defaultSearchStrategy);
        defaultSearchManager = new DefaultSearchManager(world, this, stupidSearchDecisionMaker, defaultSearchStrategy);

        pathToGo = new ArrayList<EntityID>();

        utilities.updateHealthyPoliceForceList();

        // After updating Healthy Police force information, it's time to decide which OperationMode to choose:
        operationMode = decideOperationMode();


        // In both SINGULAR_PARTITIONING and DUAL_PARTITIONING, we need to do the partitioning at Time 0 (right here) based
        // on Healthy Police Forces, namely those police forces which are already on the road.
        initializePartitionManager();


        if (MRLConstants.LAUNCH_VIEWER) {
            for (Partition partition : partitionManager.getPartitions()) {
                partition.computeConvexHull();
            }

            MrlClusterLayer.PARTITIONS_MAP.put(me().getID(), partitionManager.getPartitions());
            MrlClusterLayer.HUMAN_PARTITION_MAP.put(me().getID(), partitionManager.findHumanPartition(selfHuman));
            if (partitionManager.findHumanPartition(selfHuman) == null) {
                System.out.println(me().getID() + " I have no partition in time :" + 0);
//        }
            } else {

                System.out.println(me().getID() + " my Partition is:" + partitionManager.findHumanPartition(selfHuman).getId() + " NumberOfNeeded:" + partitionManager.findHumanPartition(selfHuman).getNumberOfNeededPFs());
            }
            MrlSubClusterLayer.Partitions = partitionManager.getPartitions();
            MrlPartitionsLayer.PARTITIONS = partitionManager.getPartitions();
            MrlRendezvousLayer.partitions = partitionManager.getPartitions();
        }

        searchHelper = new SearchHelper(world, world.getPlatoonAgent());

        System.out.println("   success");

    }

    /**
     * Creates {@link #partitionManager} and initializes it, Then sets {@link #world} partitionManager reference to the newly
     * created and initialized partition manager.
     */
    private void initializePartitionManager() {
//        if (world.getPoliceForceList().size() == 1) {
//            partitionManager = new NullPartitionManager(world);
//        } else {
        if (partitionManager == null) {
            partitionManager = new DefaultManager(world, utilities);
        }
        partitionManager.initialise();
//        }
        world.setPartitionManager(partitionManager);
    }

    /**
     * Decides which {@link OperationMode} this MrlPoliceForce should use.
     * <br/><b>Note: </b>All MrlPoliceForces should make similar decision.
     * <br/>
     * This implementation decides {@link OperationMode#SINGULAR_PARTITIONING} if number of Police Forces on the road is
     * over 66% of total Police Force count, {@link OperationMode#DUAL_PARTITIONING} otherwise.
     *
     * @return Operation mode suitable for this {@link MrlPoliceForce}
     * @see OperationMode
     * @since 1.1
     */
    private OperationMode decideOperationMode() {
        OperationMode decidedOperationMode = null;
        float healthyPoliceForcesRatio = (float) utilities.getHealthyPoliceForces().size() / (float) world.getPoliceForces().size();
        if (healthyPoliceForcesRatio > 0.66) {
            decidedOperationMode = OperationMode.SINGULAR_PARTITIONING;
        } else {
            decidedOperationMode = OperationMode.DUAL_PARTITIONING;
        }

        logger.debug("Healthy Police Forces ratio is : " + healthyPoliceForcesRatio + ", Decided Operation Mode: " + decidedOperationMode);
        return decidedOperationMode;
    }

    @Override
    protected void think(int time, ChangeSet changed, Collection<Command> heard) {
        super.think(time, changed, heard);
    }

    @Override
    public void act() throws CommandException, TimeOutException {
//      myNewAct();
        myAct();
//      oldAct();
    }

    private void myNewAct() {
//        policeBehaviorsHelper.doBehaviors();
    }

    private void oldAct() throws TimeOutException, CommandException {
        //        clearHere(CLEAR_HERE_STRATEGY);
//        loopBetweenTwoArea(world.getEntity(new EntityID(32749)), world.getEntity(new EntityID(1221)));


        world.getDecision().update();
        isThinkTimeOver("world.getDecision().update()");
//        updateUnexploredBuildings(changed);

        if ((targetPath == null && world.getTime() > 6) || world.getTime() % 10 == 0) {
            if (world.getTime() % 10 == 0) {
                world.getDecision().getTaskAssignment().clear();
            }
            targetPath = world.getDecision().assign();
            if (targetPath != null) {
                if (DEBUG_POLICE_FORCE) {
                    world.printData(" Target Selected  " + "  targetPath: " + targetPath + "  value: " + targetPath.getValue());
                }
            }
        }
        isThinkTimeOver("assign");

        clearHere(CLEAR_HERE_STRATEGY);
        isThinkTimeOver("clearHere");

//        rendezvousAction();

        if (targetPath != null) {
            selectRoadToMove();
        }
        isThinkTimeOver("selectRoadToMove");

        if (DEBUG_POLICE_FORCE) {
            world.printData(" SEARCHING ....");
        }

        search();

        //searchZones();
        isThinkTimeOver("searchZones");

        if (DEBUG_POLICE_FORCE) {
            world.printData(" REST AT REFUGE.");
        }
        restAtRefuge();
        isThinkTimeOver("restAtRefuge");
    }

    private void search() throws CommandException {
        setAgentState(State.Searching);

        heatTracerSearchManager.execute();

        possibleBuildingSearchManager.execute();
        defaultSearchManager.execute();
    }


    private void myAct() throws CommandException {

        // I'm buried and can't do anything
        if (me().getBuriedness() > 0) {
            //todo send buriedAgentMessage
            return;
        }


        try {
            utilities.updateHealthyPoliceForceList();
            partitionManager.updatePartitions(utilities.getHealthyPoliceForces().size());

            if (operationMode == OperationMode.DUAL_PARTITIONING
                    && world.getTime() == SECOND_PARTITIONING_TIME) {
                // It's about time each police force should re-calculate partitioning to include newly added healthy police
                // forces. (Whether recently buried, or healthy but inside a building)
                initializePartitionManager();
//                System.out.println(me().getID() + " my Partition is:" + partitionManager.findHumanPartition(selfHuman).getId() + " NumberOfNeeded:" + partitionManager.findHumanPartition(selfHuman).getNumberOfNeededPFs());

            }

            if (world.getTime() == 3) {

                List<StandardEntity> unKnownAgents = new ArrayList<StandardEntity>(world.getPoliceForces());
                for (Map.Entry<Human, Partition> humanPartitionEntry : partitionManager.getHumanPartitionEntrySet()) {
                    unKnownAgents.remove(humanPartitionEntry.getKey());
                }

                Human unknownHuman;
                for (StandardEntity agent : unKnownAgents) {
                    // this agent is not assigned to any partition, I will assign it to the partition it is located at.
                    unknownHuman = (Human) agent;
                    partitionManager.forceAssignAgent(unknownHuman,
                            partitionManager.findPartitionAtArea(unknownHuman.getPosition()));

                }

            } else if (partitionManager.findHumanPartition(world.getSelfHuman()) == null) {
                partitionManager.forceAssignAgent(world.getSelfHuman(),
                        partitionManager.findPartitionAtArea(selfHuman.getPosition()));
            }


            if (partitionManager.findHumanPartition(selfHuman).isDead()) {
                Partition currentPartition = partitionManager.findHumanPartition(selfHuman);
                Partition partitionToGo = utilities.findPartitionToGo(currentPartition, partitionManager.getPartitions());
                if (partitionToGo == null) {
                    partitionToGo = currentPartition;
                }
                partitionManager.forceAssignAgent(selfHuman, partitionToGo);
            }

//            utilities.updateHealthyCoworkerList(partitionManager.getCoworkers(selfHuman));

//            partitionManager.updatePartitions(utilities.getHealthyPoliceForces());

//            if(world.getTime()==15){
//                partitionManager.partitionMakingOperations();
//            }

//            if (pcc.isTimeToPartitionSplitting(world.getSegmentationHelper().getMyOriginalPartition(selfHuman), world.getSegmentationHelper().getHumanPartitionMap())) {
//
//                world.getSegmentationHelper().splitOperation(world.getSegmentationHelper().getMyOriginalPartition(selfHuman), utilities.getHealthyCoworkers());
//
            if (MRLConstants.LAUNCH_VIEWER) {
                for (Partition partition : partitionManager.getPartitions()) {
                    if (partition.getPolygon() == null) {
                        partition.computeConvexHull();
                    }
                }

                MrlClusterLayer.PARTITIONS_MAP.put(me().getID(), partitionManager.getPartitions());
                MrlClusterLayer.HUMAN_PARTITION_MAP.put(me().getID(), partitionManager.findHumanPartition(selfHuman));
                if (partitionManager.findHumanPartition(selfHuman) == null) {
                    if (partitionManager.findHumanPartition(selfHuman) == null) {
                        System.out.println(me().getID() + " I have no partition in time :" + 0);
//        }
                    } else {

                        System.out.println(me().getID() + " my Partition is:" + partitionManager.findHumanPartition(selfHuman).getId() + " NumberOfNeeded:" + partitionManager.findHumanPartition(selfHuman).getNumberOfNeededPFs());
                    }
                }
                MrlSubClusterLayer.Partitions = partitionManager.getPartitions();
                MrlPartitionsLayer.PARTITIONS = partitionManager.getPartitions();
                MrlRendezvousLayer.partitions = partitionManager.getPartitions();
            }
//            }


        } catch (Exception e) {
            logger.error("Failed to perform Partitioning related operations.");
            logger.debug("Stack Trace:", e);
        }


        if ((world.getSelfPosition() instanceof Building) && !agentState.equals(State.Searching)) {

            Building building = (Building) world.getSelfPosition();
            Point2D movePoint = null;

            List<Entrance> entrances = world.getMrlBuilding(world.getSelfPosition().getID()).getEntrances();
            if (entrances != null && !entrances.isEmpty()) {
                for (Entrance entrance : entrances) {
                    if (entrance.getNeighbour().isBlockadesDefined() && !entrance.getNeighbour().getBlockades().isEmpty()) {
                        movePoint = utilities.findMovePointInBuilding(building, entrance.getNeighbour());
                        if (movePoint != null) {
                            if (Util.distance(me().getX(), me().getY(), movePoint.getX(), movePoint.getY()) < 500) {
                                break;
                            } else {
                                moveToPoint(building.getID(), (int) movePoint.getX(), (int) movePoint.getY());
                            }
                        }
                    }
                }
            }

            clearHelper.clearBlockadesInRange(new Pair<Integer, Integer>(me().getX(), me().getY()), 1000);

            move(utilities.findNearestEntrance(entrances), 0, true);

        }

        myTask = getNextTask();

        if (myTask == null) {
            clearHelper.clearWay(null, null);
            logger.debug("getNextTask() returned null.");
        } else {
            setAgentState(State.Working);
            targetRoad = utilities.getNearestRoad(myTask.getTarget().getRoadsToMove());
            pathToGo = pathPlanner.planMove((Area) world.getSelfPosition(), targetRoad, 0, true);

            if (!pathToGo.isEmpty()) {
                clearHelper.clearWay(pathToGo, myTask.getTarget());
                pathPlanner.moveOnPlan(pathToGo);
//                sendMoveAct(world.getTime(), pathToGo);
                clearHelper.clearWay(pathToGo, myTask.getTarget());
            } else if (targetRoad != null) {
                clearHelper.clearWay(pathToGo, myTask.getTarget());
            }

            targetRoad = null;
        }

        search();
    }

    private void leaveBuilding() {
        //TODO @Pooya Assign someone to write this method.
    }

    private Task getNextTask() {
        if (strategy == null) {
            strategy = new PolicePrioritizationStrategy(world, this, inPartition, utilities, policeMessageHelper, pcc);
        }
        return strategy.act();
    }

//    TODO @Pooya Please move this useful piece of code to some document repository
//    private void act_Auctioning() {
//        if (myTask != null && pcc.isTargetCleared(myTask)) {
//            auction.doneTask();
//            System.err.println(world.getTime() + " " + me().getID() + " I have done my assigned task");
//            myTask = null;
//            targetRoad = null;
//        }
//        auction.startAuction(world.getTime(), utilities.getTargets(inPartition));
//        auction.taskAllocation(world.getTime());
//
//        myTask = auction.getTask(me().getID());
//    }

    private void updateUnexploredBuildings(ChangeSet changed) {
        if (changed == null) {
            return;
        }
        for (EntityID next : changed.getChangedEntities()) {
            unexploredBuildings.remove(next);
        }
    }

    public void clearHere(PoliceActionStyle clearType) throws CommandException {
        clearHereHelper.clear(targetRoad, clearType);
    }

    public void clearHere_new(Target target, PoliceActionStyle clearType) throws CommandException {
        clearHereHelper.clear(target, clearType);
    }

    @Override
    public void processMessage(Packet packet, EntityID sender) {
        try {
            policeMessageHelper.processMessage(packet, sender);
        } catch (Exception ex) {
            //todo: Handle it
        }

    }

    @Override
    protected EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
        return EnumSet.of(StandardEntityURN.POLICE_FORCE);
    }

    private void selectRoadToMove() throws CommandException {
        RoadHelper roadHelper = world.getHelper(RoadHelper.class);

        if (targetRoad != null && roadHelper.isPassable(targetRoad.getID())) {
            if (DEBUG_POLICE_FORCE) {
                world.printData(" target road is passable :" + targetRoad);
            }
            targetRoad = null;
        }

        if (targetPath.isEmpty()) {
            //System.out.println("self: "+world.getSelfHuman().getID()+"  time: "+world.getTime()+"  --> Error: Target Path is NULL !!!");
            return;
        }

        if (targetRoad == null) {
            //System.out.println("self: "+world.getSelfHuman().getID()+"  time: "+world.getTime()+"  value: "+targetPath.getValue()+"  targetPath: " + targetPath+"\n"+"  getImportantRoads: "+targetPath.getImportantRoads());
            for (Road road : targetPath.getImportantRoads()) {
                if (!roadHelper.isPassable(road.getID())) {
                    targetRoad = road;
                    break;
                }
            }
            if (targetRoad == null) {
                for (Road road : targetPath) {
                    if (!roadHelper.isPassable(road.getID())) {
                        targetRoad = road;
                        break;
                    }
                }
            }
            if (DEBUG_POLICE_FORCE) {
                world.printData("  targetRoad: " + targetRoad);
            }
        }
        if (targetRoad == null) {
            //System.out.println("self: "+world.getSelfHuman().getID()+"  time: "+world.getTime()+"  --> Error: Target Road is NULL !!!");
//            world.printData(" road in path set null :"+targetPath);
            targetPath = null;
            return;
        }

        if (!location().getID().equals(targetRoad.getID()) && !roadHelper.isPassable(targetRoad.getID())) {
//            if (DEBUG_POLICE_FORCE) {
//                System.out.println("         self: " + world.getSelfHuman().getID() + "  time: " + world.getTime() + "  keep moving to " + targetRoad.getID().getValue());
//            }
            if (DEBUG_POLICE_FORCE) {
                world.printData(" move to target road : " + targetRoad);
            }
            move(targetRoad, IN_TARGET, false);
        }
        if (location().getID().equals(targetRoad.getID())) {
//            System.out.println(targetPath.getTargetsInThisPath());
            for (Area area : targetPath.getImportantRoads()) {
                if (area.getNeighbours().contains(targetRoad.getID())) {
                    if (DEBUG_POLICE_FORCE) {
                        System.out.println(getDebugString() + " MOVE TO TARGET : " + area);
                    }
                    move(area, IN_TARGET, false);
                }
            }
        }
    }

    public ClearHelper getClearHelper() {
        return clearHelper;
    }
}