package mrl.world.routing.pathPlanner;

import mrl.common.CommandException;
import mrl.common.MRLConstants;
import mrl.common.Util;
import mrl.platoon.MrlPlatoonAgent;
import mrl.police.MrlPoliceForce;
import mrl.world.MrlWorld;
import mrl.world.object.MrlBuilding;
import mrl.world.object.MrlRoad;
import mrl.world.routing.a_star.A_Star;
import mrl.world.routing.graph.Graph;
import rescuecore2.log.Logger;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

/**
 * Created by Mostafa Shabani.
 * Date: Sep 22, 2010
 * Time: 4:45:00 PM
 */
public class PathPlanner implements IPathPlanner, MRLConstants {

    private MrlWorld world;
    private MrlPlatoonAgent platoonAgent;
    private A_Star a_star;
    private Graph graph;
    private CheckAreaPassable areaPassably;
    private EscapeHere escapeHere;
    public int lastMoveTime = 0;

    private List<EntityID> lastMovePlan = new ArrayList<EntityID>();
    private List<EntityID> previousPath = new ArrayList<EntityID>();
    private Area previousTarget = null;
    private Queue<EntityID> previousPositionsQueue = new LinkedList<EntityID>();
    private int continueMovingTime = 0;

    private int refugeNumberToGetNearest;
    private int nearestRefugePathCost;
    private int tryToReachTargetCount = 0;
    private EntityID prevTarget;

    public PathPlanner(MrlWorld world) {
        this.world = world;
        this.graph = new Graph(world);
        this.a_star = new A_Star(this);
        prevTarget = null;
        if (world.getSelf() instanceof MrlPlatoonAgent) {
            this.platoonAgent = world.getPlatoonAgent();
            this.areaPassably = new CheckAreaPassable(this);
            this.escapeHere = new EscapeHere(this);
        }

        // baraye inke bebinim be chand refuge bayad A* bezanim baraye be dast avardane nazdiktarin ha.
        int refSize = world.getRefuges().size();
        if ((refSize % 4) == 0) {
            refugeNumberToGetNearest = (refSize / 4);
        } else {
            refugeNumberToGetNearest = (refSize / 4) + 1;
        }
    }

    public MrlWorld getWorld() {
        return world;
    }

    @Override
    public Graph getGraph() {
        return graph;
    }

    @Override
    public CheckAreaPassable getAreaPassably() {
        return areaPassably;
    }

    @Override
    public boolean move(Area target, int maxDistance, boolean force) throws CommandException {
        setPreviousPosition();
        List<EntityID> planMove = planMove(target, maxDistance, force);
//        if (escapeHere.amIMotionLess()) {
//            EscapeState state = escapeHere.escape(target);
//            if (state.equals(EscapeState.UNREACHABLE)) {
//                setUnreachableTarget(target);
//            }
//            if (state.equals(EscapeState.TRY_NEW_A_STAR)) {
//                tryToReachTargetCount++;
//                planMove = escapeHere.newPlanMove();
//            }
//        }
//        if (planMove.isEmpty() || planMove.size() == 1) {
//            return true;
//        }

        //TODO: added to prevent moving to blocked entrances

        if (planMove.isEmpty()) {
            setUnreachableTarget(target);
            return false;
        }


        moveOnPlan(planMove);
        return false;
    }

    @Override
    public boolean move(Collection<? extends Area> targets, int maxDistance, boolean force) throws CommandException {
        setPreviousPosition();
        /*if (escapeHere.amIMotionLess() && !targets.isEmpty())
            escapeHere.escape((new ArrayList<Area>(targets)).get(0));*/
        List<EntityID> planMove = planMove(targets, maxDistance, force);
        moveOnPlan(planMove);
        return false;    // if target impassable return false.
    }

    @Override
    public void moveToRefuge() throws CommandException {
        setPreviousPosition();
        Area selfPositionArea = (Area) world.getSelfPosition();
        List<EntityID> refugePath = new ArrayList<EntityID>(getRefugePath(selfPositionArea, false));
        moveOnPlan(refugePath);
    }

    @Override
    public void moveToPoint(EntityID area, int destX, int destY) throws CommandException {

        if (world.getSelfPosition().getID().equals(area)) {
            List<EntityID> list = new ArrayList<EntityID>();
            list.add(area);
            platoonAgent.sendMoveAct(world.getTime(), list, destX, destY);
        } else {
            move((Area) world.getEntity(area), IN_TARGET, true);
        }
    }

    @Override
    public void update() {
        /**
         * dar har cycle har roadi ro ke bebine update mikone.
         * yani age grid nadashte bashe grid bandi ba'd har edgesh ro chek mikone ke aya passable hast ya na.
         */
    }

    @Override
    public List<EntityID> getLastMovePlan() {
        return lastMovePlan;
    }

    @Override
    public int getPathCost() {
        return a_star.getPathCost();
    }

    public List<EntityID> planMove(Area target, int maxDistance, boolean force) {
        if (target instanceof Building && !force) {
            if (!(platoonAgent instanceof MrlPoliceForce) && !world.getMrlBuilding(target.getID()).isOneEntranceOpen(world)) {
                return new ArrayList<EntityID>();
            }
        }

        List<EntityID> planMove = planMove((Area) world.getSelfPosition(), target, maxDistance, force);

        if (!planMove.isEmpty() && !(platoonAgent instanceof MrlPoliceForce) && !force) {
            StandardEntity entity = world.getEntity(planMove.get(planMove.size() - 1));
            boolean shouldBeRemoved = false;
            if (entity instanceof Road) {
                Road road = (Road) entity;
                if (road.isBlockadesDefined()) {
                    Blockade blockade;
                    for (EntityID blockadeId : road.getBlockades()) {
                        blockade = (Blockade) world.getEntity(blockadeId);
                        if (blockade.getShape().contains(road.getX(), road.getY())
                                || Util.findDistanceTo(blockade, road.getX(), road.getY()) <= 500) {
                            shouldBeRemoved = true;
//                            planMove.remove(road.getID());
                            break;
                        }
                    }
                }
            }
            if (shouldBeRemoved) {
                planMove = new ArrayList<EntityID>();
            }
        }


        // bedast avordane masir baraye move.
        return planMove;
    }

    @Override
    public List<EntityID> planMove(Area sourceArea, Area target, int maxDistance, boolean force) {

        List<EntityID> finalAreaPath = new ArrayList<EntityID>();

        if (target == null) {
            return finalAreaPath;
        }
        if (!target.getID().equals(prevTarget)) {
            prevTarget = target.getID();
            tryToReachTargetCount = 0;
        }
        if (sourceArea.equals(target)) {
            Logger.warn("Time:" + world.getTime() + " Already on target move =" + target);
            return finalAreaPath;
        }
        boolean repeatPlanning = repeatPlanning(target);
        boolean repeatAStar = !isPositionOnPreviousPath(sourceArea.getID());
        if (repeatAStar || repeatPlanning) {
            // tanha zamani ke niaz bashe dobare a* mizane ta masir peida kone.
            // yani vaghti yeki az sharayete bala true bashe.
            if (repeatAStar) {
                tryToReachTargetCount++;
            }
            Area nearestTarget = null;
            previousPath.clear();

            if (maxDistance != IN_TARGET) {
                // agar gharar nabood hatman dar mahale target gharar begirad.
                // ba estefade az maxDistance makan haei ke mitavanad dar aanha gharar begirad ra bedast miavarim.
                nearestTarget = getNearestArea(platoonAgent.getObjectsInRange(target, maxDistance), target);
            }
            if (nearestTarget == null) {
                nearestTarget = target;
            }

            // A* be ma te'dadi entityId mide ke marboot be area haei ke bahas agent azashoon oboor kone.
            finalAreaPath = a_star.getShortestPath(sourceArea, nearestTarget, force);

            previousTarget = nearestTarget;
            previousPath = finalAreaPath;

        } else if (previousTarget.equals(target)) {
            // baraye zamani ke meghdari az masir ra amade ast.
            // az pathe ghabli ta position alan ra hazf mikonim va pathe jadid bedast miayad.
            ArrayList<EntityID> temp = new ArrayList<EntityID>();

            for (EntityID aPreviousPath : previousPath) {
                if (!sourceArea.getID().equals(aPreviousPath)) {
                    temp.add(aPreviousPath);
                } else {
                    break;
                }
            }

            previousPath.removeAll(temp);
            finalAreaPath = previousPath;
        }
        return finalAreaPath;
    }

    @Override
    public int getNearestRefugePathCost() {
        return nearestRefugePathCost;
    }

    @Override
    public List<EntityID> getRefugePath(Area sourceArea, boolean force) {

        /**
         * aval chand ta az nazdiktarin refuge haro peyda mikonim.
         * ba'd ba hame oon ha path mizanim.
         * pathe nakdiktarin ro barmigardunim.
         */
        List<EntityID> finalPath = new ArrayList<EntityID>();

        if (world.getRefuges().size() == 0 || sourceArea instanceof Refuge) {
            // age refuge nadashte bashim ke hichi.
            return finalPath;
        }
        if (repeatPlanning(null) || !isPositionOnPreviousPath(sourceArea.getID())) {
/*
            //todo ghablan nazdiktarin refuge nazdiktarin az nazare oghlidosi entekhab mishod===MTN==>
            //agar yek refuge baste bood agentha enghadr poshte entereance montazer mimoondand ta ya refuge baz she ya
            // police uno baz kone...
            //hala aval be refuge ha A_Star zade mishe ta nazdiktarino entekhab konim bad be un path mizanim
            // agar A_Star natije nadad unvaght az nazare oghlidosi entekhab mikonim....
            Pair<EntityID, List<EntityID>> refugePathPair = getNearestRefugePath(sourceArea.getID(), false);
            List<EntityID> nearestRefugePath = refugePathPair.second();
            EntityID nearestRefuge = refugePathPair.first();
            if (nearestRefuge != null) {
                previousPath = a_star.getAreaPath(sourceArea.getID(), nearestRefuge, nearestRefugePath);
                prevTarget = nearestRefuge;
                return previousPath;
            } else {
                refugePathPair = getNearestRefugePath(sourceArea.getID(), true);
                nearestRefugePath = refugePathPair.second();
                nearestRefuge = refugePathPair.first();
                if (nearestRefuge != null) {
                    if (DEBUG_PLANNER) {
                        System.out.println("Find refuge with force");
                    }
                    previousPath = a_star.getAreaPath(sourceArea.getID(), nearestRefuge, nearestRefugePath);
                    prevTarget = nearestRefuge;
                    return previousPath;
                }
            }
            if (DEBUG_PLANNER) {
                System.out.println("a-star have no result!!!");
            }
            //A-Star natijeei dar bar nadasht hala nazdiktarin refuge az nazare oghlidosi
            //<========

*/

            List<EntityID> bestRefugePath = new ArrayList<EntityID>();
            Area bestRefuge = null;
            int minimumCost = Integer.MAX_VALUE;
            List<EntityID> path;
            int cost;
            previousPath.clear();

            List<Refuge> nearestRefuges;
            // te'dadi az refuge haye nazdik (az nazare fasele oghlidosi) ro migire.
            nearestRefuges = getSomeNearRefuge(sourceArea.getX(), sourceArea.getY());

            for (Area refuge : nearestRefuges) {
                if (refuge == null) {
                    continue;
                }

                if (sourceArea.equals(refuge)) {
                    Logger.warn("Time:" + world.getTime() + " Already on target move =" + refuge);
                    return null;
                }

                // be hameye refuge haye nazdik path mizane va nazdiktarin ro entekhab mikone.
                path = a_star.getShortestGraphPath(sourceArea, refuge, force);
                // kamtarin cost ro peyda mikonim.
                if (path.size() != 0) {
                    cost = a_star.getPathCost();

                    if (cost < minimumCost) {
                        minimumCost = cost;
                        bestRefuge = refuge;
                        bestRefugePath = path;
                    }
                } else {
                    setUnreachableTarget(refuge);
                }
            }

            if (bestRefuge == null && !nearestRefuges.isEmpty()) {
                // age refuge khubi peyda nakard avalin refuge ke az nazare oghlidosi nazdike ro entekhab mmikone.
                if (MRLConstants.DEBUG_PLANNER) {
                    world.printData(" true move....");
                }
                bestRefuge = nearestRefuges.get(0);
                bestRefugePath = a_star.getShortestGraphPath(sourceArea, bestRefuge, true);
            }
            if (bestRefuge != null) {
                if (MRLConstants.DEBUG_PLANNER) {
                    world.printData("  best:" + bestRefuge + " -------- nearest REFUGE = " + nearestRefuges);
                }
                previousTarget = bestRefuge;
                nearestRefugePathCost = minimumCost;
                previousPath = a_star.getAreaPath(sourceArea.getID(), bestRefuge.getID(), bestRefugePath);

            }

        } else {
            // age hanooz be refuge nareside bood area haei ke ta inja umade az path hazf mikone va dobare edame mide.
            ArrayList<EntityID> toRemove = new ArrayList<EntityID>();

            for (EntityID aPreviousPath : previousPath) {
                if (!sourceArea.getID().equals(aPreviousPath)) {
                    toRemove.add(aPreviousPath);
                } else {
                    break;
                }
            }
            if (MRLConstants.DEBUG_PLANNER) {
                world.printData("  continue -------- ");
            }

            previousPath.removeAll(toRemove);
        }

        finalPath = previousPath;
        return finalPath;
    }

    @Override
    public List<EntityID> getNextPlan() {
        List<EntityID> path = new ArrayList<EntityID>();
        List<EntityID> toRemove = new ArrayList<EntityID>();
        EntityID selfPos = world.getSelfPosition().getID();
        path.addAll(previousPath);

        for (EntityID aPreviousPath : previousPath) {
            if (!selfPos.equals(aPreviousPath)) {
                toRemove.add(aPreviousPath);
            } else {
                break;
            }
        }
        path.removeAll(toRemove);
        return path;
    }

    public List<EntityID> planMove(Collection<? extends Area> targets, int maxDistance, boolean force) {
        /**
         * move kardan be yek collection az target ha.
         * nazdiktarin target az nazare oghlidosi ro bedast avorde.
         * ba;d behesh path mizanim.
         */
        List<EntityID> finalAreaPath = new ArrayList<EntityID>();
        Area selfPositionArea = (Area) world.getSelfPosition();

        if (targets.size() == 0) {
            return finalAreaPath;
        }

        if (previousTarget == null || !targets.contains(previousTarget)
                || !isPositionOnPreviousPath(selfPositionArea.getID())) {
            // tanha zamani ke niaz bashe dobare a* mizane ta masir peida kone.
            // yani vaghti yeki az sharayete bala true bashe.
            Area nearestTarget = null;
            int minimumCost = Integer.MAX_VALUE;
            int cost;

            previousPath.clear();

            for (Area target : targets) {
                //peyda kardane nazdiktarin target.
                if (target == null) {
                    continue;
                }

                if (selfPositionArea.equals(target)) {
                    Logger.warn("Time:" + world.getTime() + " Already on target move =" + target);
                    return finalAreaPath;
                }
                cost = world.getDistance(target, selfPositionArea);

                if (cost < minimumCost) {
                    minimumCost = cost;
                    nearestTarget = target;
                }
            }

            // planMove previousPath ra por mikonad.
            // pas ba estefade az aan mitavan path ra bargadand.
            previousPath = planMove(selfPositionArea, nearestTarget, maxDistance, force);
            if (force && previousPath.isEmpty()) {
                /**
                 * age be in target path vojood nadashte bashe va force bashe.
                 * oon ro az targets hazf karde. va vase baghie target ha edame midahim.
                 * ta inke belakhare be yeki az oon ha path peyda konim.
                 * agar ham peyda nashod ke be jahannam.
                 */
                targets.remove(nearestTarget);
                setUnreachableTarget(nearestTarget);
                planMove(targets, maxDistance, force);
            }

        } else if (targets.contains(previousTarget)) {
            // hazfe areahei ke az aanha oboor karde ast.
            ArrayList<EntityID> temp = new ArrayList<EntityID>();

            for (EntityID aPreviousPath : previousPath) {
                if (!selfPositionArea.getID().equals(aPreviousPath)) {
                    temp.add(aPreviousPath);
                } else {
                    break;
                }
            }
            previousPath.removeAll(temp);
        }

        finalAreaPath = previousPath;
        return finalAreaPath;
    }

    @Override
    public void moveOnPlan(List<EntityID> plan) throws CommandException {
        // methodi ke act move ra be kernel ersal mikonad.

        lastMovePlan.clear();
        lastMovePlan.addAll(plan);
        if(plan.isEmpty()){
            return;
        }

        Area target = null;
        target = world.getEntity(plan.get(plan.size() - 1), Area.class);
        if (escapeHere.amIMotionLess() && !world.getPlatoonAgent().isStuck()) {
            EscapeState state = escapeHere.escape(target);
            if (state.equals(EscapeState.UNREACHABLE)) {
                setUnreachableTarget(target);
            }
            if (state.equals(EscapeState.TRY_NEW_A_STAR)) {
                tryToReachTargetCount++;
                plan = escapeHere.newPlanMove();
            }
        }
        lastMoveTime = world.getTime();
        platoonAgent.sendMoveAct(world.getTime(), plan);
    }

    private void setUnreachableTarget(Area target) {
        if (target instanceof Building) {
            MrlBuilding mrlBuilding = world.getMrlBuilding(target.getID());
            mrlBuilding.setReachable(false);
        } else {
            MrlRoad mrlRoad = world.getMrlRoad(target.getID());
            mrlRoad.setReachable(false);
        }
    }

    protected Area getNearestArea(Collection entities, Area to) {
        // entekhabe yek area ke az hame nazdiktar ast baraye zamani ke be yek collection move mizanad.

        int distance = Integer.MAX_VALUE;
        Area nearestArea = null;

        for (Object o : entities) {
            StandardEntity next = (StandardEntity) o;
            if (!(next instanceof Road)) {
                continue;
            }
            int dis = world.getDistance(to, next);
            if (dis < distance) {
                distance = dis;
                nearestArea = (Area) next;
            }
        }
        return nearestArea;
    }

    private Pair<EntityID, List<EntityID>> getNearestRefugePath(EntityID selfPosition, boolean force) {
        if (selfPosition == null)
            return null;
        Area selfPositionArea = world.getEntity(selfPosition, Area.class);

        double minDistance = Double.MAX_VALUE;
        List<EntityID> nearestPath = new ArrayList<EntityID>();
        EntityID refugeID = null;
        for (StandardEntity refugeEntity : world.getRefuges()) {
            if (world.getMrlBuilding(refugeEntity.getID()).isReachable()) {
                if (world.getPlatoonAgent() != null) {
                    Pair<Double, List<EntityID>> distancePair = world.getPlatoonAgent().getAverageTools().distancePath(selfPositionArea, world.getEntity(refugeEntity.getID(), Area.class), force);
                    double distance = distancePair.first();
                    if (distance < minDistance) {
                        minDistance = distance;
                        nearestPath = distancePair.second();
                        refugeID = refugeEntity.getID();
                    }
                }
            }
        }
        nearestRefugePathCost = (int) minDistance;
        return new Pair<EntityID, List<EntityID>>(refugeID, nearestPath);
    }

    private List<Refuge> getSomeNearRefuge(int selfX, int selfY) {
        // peyda kardane chand refuge-e nakdiz az nazare fasele oghlidosi.

        List<Refuge> nearRefuges = new ArrayList<Refuge>();
        List<Refuge> allRefuges = new ArrayList<Refuge>();
        for (StandardEntity standardEntity : world.getRefuges()) {
            Refuge refuge = (Refuge) standardEntity;
            allRefuges.add(refuge);
        }
        Refuge selected;
        Refuge absoluteNearestRefuge = null;
        int absoluteMinDist = Integer.MAX_VALUE;
        int minDistance, distance, maxDistance = -1;

        while (nearRefuges.size() < refugeNumberToGetNearest) {

            minDistance = Integer.MAX_VALUE;
            selected = null;

            for (Refuge refuge : allRefuges) {
                distance = Util.distance(selfX, selfY, refuge.getX(), refuge.getY());
                if (distance < minDistance && world.getMrlBuilding(refuge.getID()).isOneEntranceOpen(world)) {
                    minDistance = distance;
                    selected = refuge;
                    absoluteNearestRefuge = refuge;
                }
                if (absoluteMinDist > distance) {
                    absoluteMinDist = distance;
                    absoluteNearestRefuge = refuge;
                }
            }
            if (world.getSelf() instanceof MrlPoliceForce && absoluteNearestRefuge != null) {
                break;
            }
            if (maxDistance == -1) {
                maxDistance = minDistance * 2;
            } else if (maxDistance < minDistance) {
                // age distance kheili ziad beshe dige edame nemidim.
                break;
            }
            if (selected != null) {
                allRefuges.remove(selected);
                nearRefuges.add(selected);
            }
        }
        if (nearRefuges.isEmpty()) {
            nearRefuges.add(absoluteNearestRefuge);
            if (MRLConstants.DEBUG_PLANNER) {
                System.out.println("absoluteNearestRefuge: " + absoluteNearestRefuge);
            }
        }
        return nearRefuges;
    }

    private void setPreviousPosition() {
        EntityID selfPositionId = world.getSelfPosition().getID();
        if (previousPath == null || !previousPath.contains(selfPositionId)) {
            return;
        }
        for (EntityID aPreviousPath : previousPath) {
            if (!selfPositionId.equals(aPreviousPath)) {
                if (previousPositionsQueue.size() > 3) {
                    previousPositionsQueue.remove();
                }
                previousPositionsQueue.add(aPreviousPath);
            } else {
                break;
            }
        }
    }

    public EntityID getPreviousPosition() {
        if (!previousPositionsQueue.isEmpty()) {
            return previousPositionsQueue.remove();
        }
        return null;
    }

    private boolean isPositionOnPreviousPath(EntityID position) {
        return previousPath.contains(position);
    }

    private boolean repeatPlanning(Area target) {
        if (previousTarget == null || !previousTarget.equals(target) || continueMovingTime > 4) {
            continueMovingTime = 0;
            return true;
        } else {
            continueMovingTime++;
            return false;
        }
    }

    /**
     * be ma mige ke che tedad talash baraye residan be 1 target khas az masirhaye motefavet anjam shode
     * (jame tedad A_Star haei ke be vaseteye baste boodane masir ya baz shodane masire jadid zade shod)
     *
     * @return try to reach one target
     */
    public int tryToReachTargetCount() {
        return tryToReachTargetCount;
    }
}